import os
import json
import inspect
from datetime import timedelta, datetime
import requests
from concurrent.futures import ThreadPoolExecutor
from typing import List

import pandas as pd
import numpy as np
from simple_salesforce import Salesforce


from ..time import TimeIt


filedir = os.path.abspath(os.path.dirname(__file__))


class SalesforceObj():
    
    TYPES_MAPPING = {
        'string': str,
        'reference': str,
        'datetime': 'datetime',
        'picklist': str,
        'boolean': bool,
        'date': 'date',
        'double': float,
        'phone': str,
        'url': str,
        'email': str,
        'id': str,
        'textarea': str,
        'int': int,
        'address': str,
        'multipicklist': list,
        'currency': float,
        'percent': float
    }

    MAX_QUERY_SIZE = 100

    def __init__(
        self,
        config,
        client_id=None,
        client_secret=None,
        ):
        req_params =  inspect.getargspec(Salesforce)[0]
        self.config_params = {i: j for i, j in config.items() if i in req_params}
        self.config_params['version'] = '53.0'

        self.client_id = client_id
        self.client_secret = client_secret
        self.load_client_credentials()
        
        self.access_token = None
        self.instance_url = None
        self.login()

    def load_client_credentials(self):
        if not self.client_id or not self.client_secret:

            with open(os.path.join(filedir, 'credentials.json'), 'r') as f:
                file = f.read()

            try:
                credentials = json.loads(file)
                self.client_id = credentials.get('CLIENT_ID')
                self.client_secret = credentials.get('CLIENT_SECRET')

            except Exception:
                print('Could not load SF client credentials.')

    def login(self):
        '''
        This class will eventually only use oauth2 refresh tokens because
        starting in Feb 2022, all username/password login will require MFAs

        '''
        refresh_token = self.config_params.get('session_id') 

        if refresh_token:
            print('Logging in with refresh token')
            self.refresh_token(refresh_token)
            self.sf = Salesforce(instance_url=self.instance_url, session_id=self.access_token)

        else:
            print('Logging in with username/password.')
            self.sf = Salesforce(**self.config_params)

    def refresh_token(
        self,
        refresh_token: str,
        client_id: str=None,
        client_secret: str=None,
        url="https://login.salesforce.com/services/oauth2/token",
        ):
        '''
        Refresh access token for login.

        Args:
            - refresh_token (str)

            - client_id (str, default=None)

            - client_secret (str, default=None)

            - url (str, default='https://login.salesforce.com/services/oauth2/token')
            
        '''
        payload = {
            'grant_type': 'refresh_token',
            'client_id': client_id or self.client_id,
            'client_secret': client_secret or self.client_secret,
            'refresh_token': refresh_token
        }
        resp = requests.post(url=url, data=payload).json()
        self.access_token = resp['access_token']
        self.instance_url = resp['instance_url']
            
    def check_table_exists(
        self,
        tablename,
        verbose=True
        ):
        '''
        Custom sobjects/tables are not included in the property attribute,
        hence we use this wrapper to see if the sobject exits for the client.
        
        Args:
            tablename (str): name of the sobject.

            verbose (bool, default=False): print the possible error.
        '''
        try:
            query = '''
                SELECT
                    COUNT(Id)
                FROM
                    {}
            '''.format(tablename)
            count = self.sf.query(query).get('records')[0].get('expr0')

            if count == 0:
                return None

            return True

        except Exception as e:
            if verbose:
                print(e)

            return None

    @property
    def tables(self):
        '''
        Get all the tables/SObjects
        '''
        return [obj['name'] for obj in self.sf.describe()['sobjects']]

    def get_table_info(self, tablename, columns=None):
        if not self.check_table_exists(tablename):
            return None

        table = getattr(self.sf, tablename)
        df = pd.DataFrame(table.describe()['fields'])

        if columns is not None and all(item in df['name'].tolist() for item in columns):
            df = df[columns]

        return df

    def map_types(
        self,
        df,
        tablename,
        check_column_casing=True,
        return_as_dict=False
    ):
        '''
        Preparation of dataframe to be Salesforce compatible.

        Args:
            - df (pd.DataFrame)

            - tablename (str): Sobject in Salesforce.

            - check_column_casing (bool, default=True): Manage casing for column
                names.

            - return_as_dict (bool, default=False): Return as dataframe or dict.

        Returns:
                - df or data_list 

        '''
        df = df.copy()
        
        info = self.get_table_info(tablename)[['name', 'type']]

        if check_column_casing:
            info['lower_name'] = info['name'].str.lower()
            case_mapping = {i: j for i, j in info.set_index('lower_name')['name'].to_dict().items() if i in df.columns} 
            df.rename(columns=case_mapping, inplace=True)

        info = info.loc[info.name.isin(df.columns)]

        info['python_type'] = info['type'].map(type(self).TYPES_MAPPING)
        info.loc[info['python_type'].isnull(), 'python_type'] = str

        mapping = info.set_index('name')['python_type'].to_dict()
        mapping = {i: j for i, j in mapping.items() if j not in [list, dict]}

        df = df[[col for col in df.columns if col in mapping.keys()]]

        for k, v in mapping.items():

            if v in [list, dict]:
                continue

            elif v == "date":
                df[k] = df[k].astype('datetime64[ns]').dt.strftime("%Y-%m-%d")

                if return_as_dict:
                    df[k] = df[k].fillna(-99_999_999)

            elif v == 'datetime':
                df[k] = df[k].astype('datetime64[ns]').dt.strftime('%Y-%m-%dT00:00:00.000Z')

                if return_as_dict:
                    df[k] = df[k].fillna(-99_999_999)

            else:
                df[k] = df[k].astype(v, errors='ignore')

            # Trick to handle null value for numeric when returning dict
            if v in [int, float] and return_as_dict is True:
                df[k] = df[k].fillna(-99_999_999)

        df = df.where(pd.notnull(df), None)

        if return_as_dict is False:
            return df

        # Returning as dict handling nan values
        data = df.to_dict(orient='records')

        data_list = []

        for d in data:
            new = {
                i: (None if j in [
                    pd.NaT,
                    np.nan,
                    'nan',
                    -99_999_999
                ] else j) for i, j in d.items()
            }

            data_list.append(new)

        return data_list
        
    def get_table_cols(
        self,
        tablename
        ):
        '''
        Returns all column names of table / SObject of self.sf

        Args:
            tablename (str)

        Returns
            Column names    
        '''
        return self.get_table_info(tablename, columns=['name'])['name'].tolist()

    @TimeIt()
    def query(
        self,
        tablename,
        columns=None,
        limit=None,
        df=False,
        date_from=None,
        date_to=None,
        date_window=None,
        date_window_variable='LastModifiedDate',
        verbose=False,
        timeout=None,
        max_columns=100, # SF limit on columns
        ):
        '''
        Returns dictionary of the columns requested from the table / SObject specified.
        If no columns or wrong columns are provided, it returns all the columns.

        Args:
            tablename (str): SObject Resource from self.sf
            
            columns (str: default=None): Columns names to be queried.

            limit (int): CHECK THIS! You have to add a limit.

            df(boolean default: False): Output format Dict or pd.DataFrame
            
            date_from (str, default=None): It has to be in the correct format.

            date_to (str, default=None): It has to be in the correct format.
             
            date_window (int, default=None): How many days to go back
             
            verbose (bool, default=False): print the query

        Returns:
            Output format Dict or pd.DataFrame from query

        Notes:
            - You have to decide between date_window or (date_to and date_from)
                There is no validation for this observation.

        '''
        ####
        def chunker(seq, size):
            return (seq[pos:pos + size] for pos in range(0, len(seq), size))
        #####

        if not hasattr(self.sf, tablename):
            return None

        column_names = self.get_table_cols(tablename)

        if columns is None or not all(item in column_names for item in columns):
            columns = column_names

        # Splitting in chunks
        chunks = chunker(columns, max_columns - 1)

        counter = 0
        for chunk in chunks:

            if 'Id' not in chunk:
                chunk = ['Id'] + chunk
            query_cols = ', '.join(chunk)
            query = '''
                SELECT
                    {}
                FROM
                    {}
            '''.format(
                query_cols,
                tablename
            )
            # Using date_window
            if isinstance(date_window, int) and date_window > 0:
                date_from = (
                    datetime.today() - timedelta(days=date_window)
                ).strftime('%Y-%m-%dT00:00:00.000Z')

                query += f" WHERE {date_window_variable} > {date_from}"

            # Using date_to AND date_from
            elif date_from and date_to:
                query += f" WHERE {date_window_variable} > {date_from} AND {date_window_variable} < {date_to}"

            elif date_from:
                query += f" WHERE {date_window_variable} > {date_from}"

            elif date_to:
                query += f" WHERE {date_window_variable} < {date_to}"

            if isinstance(limit, int):
                query += f' LIMIT {limit}'

            if verbose:
                print(query)

            results = self.sf.query_all(query, include_deleted=True, timeout=timeout)['records'] 

            # Only easy way to marge chunks on Id through a DataFrame
            results = pd.DataFrame.from_dict(results)

            if results.shape[0] < 1:
                return None

            if 'attributes' in results.columns:
                results.drop('attributes', axis=1, inplace=True)
            else:
                for r in results:
                    r.pop('attributes', None)

            if counter == 0:
                all_results = results
            
            else:
                all_results = pd.merge(all_results, results, on='Id', how='left')

            counter += 1            

        if df == False:
            all_results = all_results.to_dict(orient='records')

        return all_results

    ############ PARALLEL QUERY ##########
    @TimeIt()
    def query_parallel(
        self,
        tablename,
        columns=None,
        limit=None,
        df=False,
        date_window=None,
        date_window_variable='LastModifiedDate',
        verbose=False,
        n_chunks=4 # Unfortunately SF limits this to 10
    ):
        ###### 
        def limit_split(limit, n):
            if limit is None:
                return [None] * n
            l = [int(limit / n)] * n
            diff = limit % n
            for i in range(diff):
                l[i] += 1
            return l
        ############

        if date_window:
            date_from = (datetime.today() - timedelta(days=date_window))
        
        else:
            # If no date_window is provided, we retrive the oldest date according to "CreatedDate"
            date_window_variable = 'CreatedDate'

            min_date_query = f"SELECT MIN({date_window_variable}) FROM {tablename}"
            date_from_fmt = self.sf.query(min_date_query, include_deleted=True)['records'][0]['expr0']
            date_from = datetime.strptime(date_from_fmt, '%Y-%m-%dT%H:%M:%S.000+0000')

        date_from_fmt = date_from.strftime('%Y-%m-%dT00:00:00.000Z')

        date_to = datetime.today() + timedelta(days=1)

        diff = (date_to  - date_from ) / n_chunks

        chunks = []

        for i in range(n_chunks):
            start = (date_from + diff * i).strftime('%Y-%m-%dT00:00:00.000Z')
            end = (date_from + diff * (i + 1)).strftime('%Y-%m-%dT00:00:00.000Z')
            chunks.append((start, end))
        
        chunks = [c for c in chunks if c[0] != c[1]]

        payload = {
            'tablename': [tablename] * len(chunks),
            'columns': [columns] * len(chunks),
            'limit': limit_split(limit, n_chunks),
            'df': [df] * len(chunks),
            'date_from': [i[0] for i in chunks],
            'date_to': [i[1] for i in chunks],
            'data_window': [date_window_variable] * len(chunks),
            'data_window_variable': [date_window_variable] * len(chunks),
            'verbose': [verbose] * len(chunks)
        }

        with ThreadPoolExecutor() as executor:
            results = executor.map(self.query, *payload.values())
        results = [*results]
        
        if df:
            results = pd.concat(results).reset_index(drop=True)

        else:
            results = [item for sublist in results for item in sublist]
        
        return results
    ######################################

    def update(
        self,
        tablename,
        record_list,
        batch_size=10000
        ):
        '''
        Args:
            tablename (str): name of the sobject.
            record_list (list or np.array or pd.Series): List
                of records for the update
            batch_size (int, default=100000): After 10_000, it uses multithreading.

            format: "[{'Id': "11111", 'field1': '1111'}]"
        '''
        if not self.check_table_exists(tablename):
            return None

        bulk_handler = getattr(self.sf, "bulk")
        bulk_object = getattr(bulk_handler,  tablename)

        result = bulk_object.update(data=record_list, batch_size=batch_size)

        return result

    def insert_df(
        self,
        tablename: str,
        dataframe: pd.DataFrame,
        batch_size: int=10_000,
        update=True,
        insert=True,
        conflict_on: str or List[str] ='Name',
        return_response=False,
    ):
        '''
        Args:
            - tablename (str): SObject in Salesforce

            - dataframe (pd.DataFrame): Data dataframe.

            - batch_size (int, default=10_000): Salesforce default.

            - update (bool, default=True): Update records that already exist.

            - insert (bool, default=True): Insert new records.

            - conflict_on (str or list, default='Name'): Specify unique column(s) constrain
                to separate between update and insert.

                ** These are case sensitive and have to match the Salesforce fields.

                ** Numeric types are not supported for this argument.

            - return_response (bool, default=False): Include in response_payload
                the raw response from the Salesforce Api.

        Returns:
            - response_payload (dict)

        '''
        if not self.check_table_exists(tablename):
            return None

        dataframe = dataframe.copy()

        conflict_flag = True if isinstance(conflict_on, list) else False

        # STEP 1: Separate new from existing according to "conflict_on"

        # Query logic for multi-value conlict_on
        if conflict_flag:

            for col in conflict_on:
                dataframe[col] = dataframe[col].str.replace("'", "\\'")

            conflict_on_list = dataframe[conflict_on].to_dict(orient='records')

            conflict_on_list_fmt = []

            for i in conflict_on_list:
                fmt_list = []

                for k, j in i.items():
                   fmt_list.append(f"{k} = '{j}' ") 

                fmt_list = "(" + "AND ".join(fmt_list) + ")"
                conflict_on_list_fmt.append(fmt_list)

            conflict_on_list = conflict_on_list_fmt

        else:
            # For simple conflict on.
            conflict_on_list = dataframe[conflict_on].str.replace("'", "\\'").tolist()

        # Query as a maximum for "WHERE IN"
        def chunker(seq, size):
            return (seq[pos:pos + size] for pos in range(0, len(seq), size))

        chunks = chunker(
            conflict_on_list,
            type(self).MAX_QUERY_SIZE
        )

        all_results = []

        for chunk in chunks:

            if conflict_flag:
                conflict_on_list = ' OR '.join(chunk)

            else:
                conflict_on_list = ','.join([f"'{i}'" for i in chunk])

            query = f'''
                SELECT
                    Id,
                    {', '.join(conflict_on) if conflict_flag else conflict_on}
                FROM
                    {tablename}
                WHERE
                    isdeleted = FALSE
            '''

            if conflict_flag:
                query += f" AND ({conflict_on_list})"

            else:
                query += f" AND {conflict_on} IN ({conflict_on_list}) "

            results = self.sf.query_all(
                query,
                include_deleted=True,
                timeout=None
            )['records']

            results = pd.DataFrame.from_dict(results)
            
            all_results.append(results)

        results = pd.concat(all_results, axis=0, ignore_index=True)

        columns = ["Id"] + conflict_on if conflict_flag else ["Id", conflict_on]

        if results.shape[0] > 0:
            results_mapping =  results[columns]

        else:
            results_mapping = pd.DataFrame(columns=columns)

        dataframe = pd.merge(
            dataframe,
            results_mapping,
            on=conflict_on,
            how='left'
        )

        existing_df = dataframe[
            ~dataframe['Id'].isnull()
        ]

        print('Existing records: ', existing_df.shape[0])

        new_df = dataframe[
            dataframe['Id'].isnull()
        ].drop('Id', axis=1)

        print('New records: ', new_df.shape[0])

        # Settings objects to update and insert
        bulk_handler = getattr(self.sf, "bulk")
        bulk_object = getattr(bulk_handler,  tablename)

        response_payload = {}

        # Existing records
        if update:
            existing_data = self.map_types(
                df=existing_df,
                tablename=tablename,
                return_as_dict=True
            )

            existing_results = bulk_object.update(
                data=existing_data,
                batch_size=batch_size
            )

            conflict_on_list = conflict_on if conflict_flag else [conflict_on]

            for i, j in zip(existing_data, existing_results):

                for c in conflict_on_list:
                    j[c] = i[c]

            response_payload['update'] = {'success': 0, 'failure': 0}

            for i in existing_results:
                if i['success'] == True:
                    response_payload['update']['success'] += 1

                elif i['success'] == False:
                    response_payload['update']['failure'] += 1

            if return_response:
                response_payload['update']['results'] = existing_results

        # New records
        if insert:
            new_data = self.map_types(
                df=new_df,
                tablename=tablename,
                return_as_dict=True
            )
            
            new_results = bulk_object.insert(
                data=new_data,
                batch_size=batch_size
            )

            conflict_on_list = conflict_on if conflict_flag else [conflict_on]

            for i, j in zip(new_data, new_results):
                for c in conflict_on_list:
                    j[c] = i[c]

            response_payload['insert'] = {'success': 0, 'failure': 0}

            for i in new_results:
                if i['success'] == True:
                    response_payload['insert']['success'] += 1

                elif i['success'] == False:
                    response_payload['insert']['failure'] += 1

            if return_response:
                response_payload['insert']['results'] = new_results

        return response_payload
